/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
package ca.uvic.csr.shrimp.DisplayBean.NodeShape;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Shape;
import java.awt.geom.Rectangle2D;
import java.awt.image.ImageObserver;

import javax.swing.JPanel;


/**
 * Base class for {@link NodeShape} objects.
 * 
 * Can also render images on top of the nodes.  These images can be scaled in different ways:
 * <dl>
 * <dt>Tiled:</dt>
 * <dd>Tiles the image over the whole node.  If the image is larger than the node
 * then the image is cutoff.</dd>
 * 
 * <dt>Stretched:</dt>
 * <dd>Stretches or shrinks the image to always fit the size of the node.</dd>
 * 
 * <dt>Centered:</dt>
 * <dd>Centers the image on the node and draws it.  If the image is larger than the
 * node then it will be shrunk down to match the node size.</dd> 
 * 
 * <dt>Normal:</dt>
 * <dd>Draws the image in the top left with no scaling.  If the image is larger than 
 * the node then it is clipped. </dd>
 * </dl>
 
 * 
 * @author Rob Lintern
 */
public abstract class AbstractNodeShape implements NodeShape, ImageObserver {

	protected String name;
	private int tilePadding = 0;
	private boolean displayNoImageMessage = false;	// not used
	
	private NodeImage customImage = null;

	protected AbstractNodeShape(String name) {
		this.name = name;
	}
	
	/////////////////////////////////
	// Extended/Implemented methods
	/////////////////////////////////

	public boolean imageUpdate(Image img, int infoflags, int x, int y, int width, int height) {
		return true;
	}
	
	/**
	 * @see ca.uvic.csr.shrimp.DisplayBean.NodeShape.NodeShape#clone()
	 */
	public abstract Object clone();
	
	/**
	 * @see ca.uvic.csr.shrimp.DisplayBean.NodeShape.NodeShape#getName()
	 */
	public String getName() {
		return name;
	}

	public String toString() {
		return name;
	}
	
	/**
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Object o) {
		return this.getName().compareTo(((NodeShape)o).getName());
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {
		if (obj instanceof NodeShape) {
			return this.getName().equals(((NodeShape)obj).getName());
		}
		return false;
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		return getName().hashCode();
	}
	
	/**
	 * Sets the custom image, fill, and draw border properties for this node shape.
	 */
	public void setCustomRendering(NodeImage nodeImage) {
		this.customImage = nodeImage;
	}
	
	public boolean hasCustomRendering() {
		return (customImage != null);
	}
	
	///////////////////////////////
	// RENDER METHODS
	///////////////////////////////
		
	/** 
	 * @see ca.uvic.csr.shrimp.DisplayBean.NodeShape.NodeShape#render(Rectangle2D, Graphics2D, Color, Color, float)
	 */
	public void render(Rectangle2D outerBounds, Graphics2D g2, Color fillColor, 
			Color strokeColor, float strokeWidth, NodeImage nodeImage) {
		
	    //double r = Math.max(outerBounds.getWidth(), outerBounds.getHeight());
		//g2.setPaint(new RoundGradientPaint (outerBounds.getCenterX(), outerBounds.getCenterY(), fillColor, new Point2D.Double(0, r), Color.BLACK));
		
		// override and use the custom settings
		if (customImage != null) {
			nodeImage = customImage;
		}
				
		if (nodeImage.isFillBackground()) {
			g2.setPaint(fillColor);
			renderShape(outerBounds, g2);
		}
		
		Image image = nodeImage.getImage();
		if (image != null) {		
			renderImage(outerBounds, g2, image, nodeImage.getImageSizing());
		}

		if (nodeImage.isDrawBorder()) {
			g2.setPaint(strokeColor);
			g2.setStroke(new BasicStroke(strokeWidth));
			renderBorder(outerBounds, g2, strokeWidth);
		}
		
		renderOverlay(outerBounds, g2, fillColor, strokeColor, strokeWidth, nodeImage);
	}
	
	/**
	 * Allows subclasses to render something over top of the node (over the node shape, image, and border). 
	 */
	protected void renderOverlay(Rectangle2D outerBounds, Graphics2D g2, Color fillColor, 
			Color strokeColor, float strokeWidth, NodeImage nodeImage) {
		// does nothing
	}
	
	/**
	 * Fills the shape.
	 * @param outerBounds
	 * @param g2
	 */
	protected void renderShape(Rectangle2D outerBounds, Graphics2D g2) {
		Shape shapeToRender = getShape(outerBounds);
		g2.fill(shapeToRender);
	}
	
	/**
	 * @param g2
	 * @param strokeWidth
	 */
	protected void renderBorder(Rectangle2D outerBounds, Graphics2D g2, float strokeWidth) {
		double inset = strokeWidth/2.0;
		double x = outerBounds.getX() + inset;
		double y = outerBounds.getY() + inset;
		double w = outerBounds.getWidth() - 2*inset;
		double h = outerBounds.getHeight() - 2*inset;
		Rectangle2D boundsWithInset = new Rectangle2D.Double (x, y, w, h);
		g2.draw(getShape(boundsWithInset));
	}
	
	/**
	 * @param outerBounds
	 * @param g2
	 */
	protected void renderImage(Rectangle2D outerBounds, Graphics2D g2, Image image, String imageSizingMode) {
		if (image != null) {
			int imageWidth = image.getWidth(this);
			int imageHeight = image.getHeight(this);
			if ((imageWidth > 0) && (imageHeight > 0)) {
				if (NodeImage.STRETCHED.equals(imageSizingMode)) {
					drawStretchedImage(outerBounds, g2, image);
				} else if (NodeImage.CENTERED.equals(imageSizingMode)) {
					drawCenteredImage(outerBounds, g2, image, imageWidth, imageHeight, true);
				} else if (NodeImage.TILED.equals(imageSizingMode)) {
					drawTiledImage(outerBounds, g2, image, imageWidth, imageHeight);
				} else {
					drawNormalImage(outerBounds, g2, image, imageWidth, imageHeight);
				}
			}
		} else if (displayNoImageMessage ) {
			drawNoImageString(outerBounds, g2);
		}
	}

	/**
	 * Writes out "No Image" in the bottom right corner of the node.
	 */
	private void drawNoImageString(Rectangle2D outerBounds, Graphics2D g2) {
		// write out a string when no image is found
		String msg = "No Image";
		// save the original font and color
		Font oldFont = g2.getFont();
		g2.setFont(oldFont.deriveFont(24f));
		Color oldColor = g2.getColor();
		g2.setColor(Color.gray);
		int width = g2.getFontMetrics().stringWidth(msg) + 24;
		int height = g2.getFontMetrics().getHeight(); 
		if ((width < outerBounds.getWidth()) && (height < (2 * outerBounds.getHeight()))) {
			// align bottom right
			float x = (float)(outerBounds.getMaxX() - width);
			float y = (float)(outerBounds.getMaxY() - height);
			g2.drawString(msg, x, y);
		}
		// restore the original font and color
		g2.setColor(oldColor);
		g2.setFont(oldFont);
	}	
	
	private void drawCenteredImage(Rectangle2D bounds, Graphics2D g2, Image image, int imageWidth, int imageHeight, boolean shrink) {
		int x = (int) bounds.getX();
		int y = (int) bounds.getY();
		int w = Math.min(imageWidth, (int)bounds.getWidth());
		int h = Math.min(imageHeight, (int)bounds.getHeight());
		
		if (shrink && ((imageWidth > bounds.getWidth()) || (imageHeight > bounds.getHeight()))) {
			// shrink image to match bounds - keep aspect ratio fixed
			double scaleWidth = bounds.getWidth() / imageWidth;
			double scaleHeight = bounds.getHeight() / imageHeight;
			if (scaleWidth < scaleHeight) {
				w = (int)(scaleWidth * imageWidth);
				h = (int)(scaleWidth * imageHeight);
			} else {
				w = (int)(scaleHeight * imageWidth);
				h = (int)(scaleHeight * imageHeight);
			}
			// scales the image
			g2.drawImage(image, x, y, w, h, /* g2.getColor(),*/ this);
		} else {
			// center the image
			x += (int)Math.max(bounds.getX(), (bounds.getWidth() - w) / 2);	 // center horizontally
			y += (int)Math.max(bounds.getY(), (bounds.getHeight() - h) / 2); // center vertically
			// chop off the outer rim of the image if necessary
			int srcX = Math.max(0, (imageWidth / 2) - (w / 2));
			int srcY = Math.max(0, (imageHeight / 2) - (h / 2));
			// no scaling is done here - note it uses only (x, y) coordinates, not width and height
			g2.drawImage(image, x, y, x+w, y+h, srcX, srcY, srcX+w, srcY+h, /* g2.getColor(),*/ this);
		}
	}

	private void drawNormalImage(Rectangle2D bounds, Graphics2D g2, Image image, int imageWidth, int imageHeight) {
		int x = (int) bounds.getX();
		int y = (int) bounds.getY();
		int w = Math.min(imageWidth, (int)bounds.getWidth());
		int h = Math.min(imageHeight, (int)bounds.getHeight());
		
		// no scaling is done here - note it uses only (x, y) coordinates, not width and height
		g2.drawImage(image, x, y, x+w, y+h, 0, 0, w, h, /* g2.getColor(),*/ this);
	}
	
	private void drawStretchedImage(Rectangle2D bounds, Graphics2D g2, Image image) {
		int x = (int) bounds.getX();
		int y = (int) bounds.getY();
		int w = (int) bounds.getWidth();
		int h = (int) bounds.getHeight();
		// scales the image
		g2.drawImage(image, x, y, w, h, this);
	}
	
	private void drawTiledImage(Rectangle2D bounds, Graphics2D g2, Image image, int imageWidth, int imageHeight) {
		int x = (int) bounds.getX();
		int y = (int) bounds.getY();
		int w = imageWidth;
		int h = imageHeight;
		
		int cols = (int) Math.ceil(bounds.getWidth() / imageWidth);
		int rows = (int) Math.ceil(bounds.getHeight() / imageHeight);
		for (int row = 0; row < rows; row++) {
			if ((y + imageHeight) > bounds.getMaxY()) {
				h = (int) bounds.getMaxY() - y;
			}
			for (int col = 0; col < cols; col++) {
				w = imageWidth;
				if ((x + imageWidth) > bounds.getMaxX()) {
					w = (int) bounds.getMaxX() - x;
				}
				// no scaling
				g2.drawImage(image, x, y, x + w, y + h, 0, 0, w, h, this);
				// move over a column
				x += w + tilePadding;
			}
			// go down a row
			y += h + tilePadding;
			x = (int) bounds.getX();
		}
	}	

	/** 
	 * @see ca.uvic.csr.shrimp.DisplayBean.NodeShape.NodeShape#getThumbnail(int, int, int, int, Color)
	 */
	public JPanel getThumbnail(int x, int y, int width, int height, Color color, NodeImage nodeImage) {
		final NodeShape thumbShape = (NodeShape) this.clone();
		thumbShape.setCustomRendering(customImage);
		JPanel thumbPanel = new ThumbPanel(thumbShape, color, nodeImage, new Rectangle2D.Double(x, y, width, height));
		thumbPanel.setPreferredSize(new Dimension(width, height));
		return thumbPanel;
	}

	class ThumbPanel extends JPanel {
		
		private Color fillColor;
		private NodeShape nodeShape;
		private Rectangle2D outerBounds;
		private NodeImage nodeImage;
		
		public ThumbPanel (NodeShape nodeShape, Color fillColor, NodeImage nodeImage, Rectangle2D outerBounds) {
			super ();
			this.nodeShape = nodeShape;
			this.fillColor = fillColor;
			this.nodeImage = nodeImage;
			this.outerBounds = outerBounds;
		}
			
		public void paint(Graphics g) {
			super.paint(g);
			Graphics2D g2= (Graphics2D)g;
			nodeShape.render(outerBounds, g2, fillColor, Color.black, 1.0f, nodeImage);
		}			
	}

	/**
	 * Always returns true.  Subclasses that use custom node shapes should return false 
	 * if users shouldn't be allowed to select their node shape. 
	 * @see NodeShape#isUserSelectable()
	 */
	public boolean isUserSelectable() {
		return true;
	}

	/** 
	 * Subclasses override this.
	 * This determines whether the inner border should be drawn. 
	 */
	public boolean drawInnerBorder() {
		return false;
	}
	
}
